// Implementation of the RoundedCube class.
// Copyright (C) 2002  Martin Geisler <gimpster@gimpster.com>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "RoundedCube.h"

void RoundedCube::CopyColor(float a[], float b[]) {
  a[0] = b[0];
  a[1] = b[1];
  a[2] = b[2];
}

void RoundedCube::RotateColors(int a, int b, int c, int d) {
  float tmp[3];

  CopyColor(tmp, sides[a]);
  CopyColor(sides[a], sides[b]);
  CopyColor(sides[b], sides[c]);
  CopyColor(sides[c], sides[d]);
  CopyColor(sides[d], tmp);
}

RoundedCube::RoundedCube(float s_length, float r_width, int r_steps) {
  draw_normals = false;
  side_length = s_length;
  round_steps = r_steps;
  round_width = r_width;

  // Original colors:
  sides[0][0] = 0; sides[0][1] = 1;   sides[0][2] = 0; // Green
  sides[1][0] = 1; sides[1][1] = 1;   sides[1][2] = 1; // White
  sides[2][0] = 1; sides[2][1] = 0.5; sides[2][2] = 0; // Orange
  sides[3][0] = 1; sides[3][1] = 0;   sides[3][2] = 0; // Red
  sides[4][0] = 1; sides[4][1] = 1;   sides[4][2] = 0; // Yellow
  sides[5][0] = 0; sides[5][1] = 0;   sides[5][2] = 1; // Blue

  sides_specular[0] = sides_specular[1] = sides_specular[2] = 0.75;
  sides_specular[3] = 1.0;
  sides_shininess[0] = 16; // The sides have small highlights.

  bevel_diffuse[0] = bevel_diffuse[1] = bevel_diffuse[2] = 0.25;
  bevel_diffuse[3] = 1.0;
  bevel_specular[0] = bevel_specular[1] = bevel_specular[2] = 1.0;
  bevel_specular[3] = 1.0;
  bevel_shininess[0] = 128;
}

void RoundedCube::rotate(axis a, bool direction) {

  switch (a) {
  case Xaxis:
    if (direction)
      RotateColors(3, 4, 2, 1);
    else
      RotateColors(1, 2, 4, 3);
    break;
  case Yaxis:
    if (direction)
      RotateColors(0, 2, 5, 3);
    else
      RotateColors(3, 5, 2, 0);
    break;
  case Zaxis:
    if (direction)
      RotateColors(0, 4, 5, 1);
    else
      RotateColors(1, 5, 4, 0);
    break;
  default:
    ;
  }

}


void RoundedCube::DrawMarker() {

  glBegin(GL_TRIANGLES);

  glColor3f(1, 0, 0);
  glVertex3f(-0.2, -0.2, -0.5);
  glVertex3f( 0.2, -0.2, -0.5);
  glVertex3f(   0,  0.2, -0.5);

  glVertex3f(-0.2, -0.2, -0.5);
  glVertex3f( 0.2, -0.2, -0.5);
  glColor3f(1, 0, 0);
  glVertex3f(0, 0, 0);
  glColor3f(0, 1, 0);

  glVertex3f( 0.2, -0.2, -0.5);
  glVertex3f(   0,  0.2, -0.5);
  glColor3f(1, 0, 0);
  glVertex3f(0, 0, 0);
  glColor3f(0, 1, 0);

  glVertex3f(-0.2, -0.2, -0.5);
  glVertex3f(   0,  0.2, -0.5);
  glColor3f(1, 0, 0);
  glVertex3f(0, 0, 0);
  glColor3f(0, 1, 0);

  glEnd();

}

void RoundedCube::DrawArrow(float x1, float y1, float z1,
                            float x2, float y2, float z2) {
  
  glBegin(GL_LINES);
  glColor3f(1, 0, 0);
  glVertex3f(x1, y1, z1);
  glColor3f(0, 1, 0);
  glVertex3f(x2, y2, z2);
  glEnd();
  
}           



void RoundedCube::MakeCornerSphere() {
  
  float step_size = M_PI/2/round_steps;
  float v, w, vv, ww;

  glBegin(GL_QUADS);
  glColor3f(0.5, 0.5, 0.5);
  
  for (int i = 0; i < round_steps; i++) {
    v = i * step_size;
    vv = (i+1)*step_size;

    for (int j = 0; j < round_steps; j++) {
      w = j * step_size;
      ww = (j+1) * step_size;

      glNormal3f(cos(v) * cos(w), cos(v) * sin(w), sin(v));
      glVertex3f(cos(v) * cos(w) * round_width,
                 cos(v) * sin(w) * round_width,
                 sin(v) * round_width);

      glNormal3f(cos(vv) * cos(w), cos(vv) * sin(w), sin(vv));
      glVertex3f(cos(vv) * cos(w) * round_width,
                 cos(vv) * sin(w) * round_width,
                 sin(vv) * round_width);

      glNormal3f(cos(vv) * cos(ww), cos(vv) * sin(ww), sin(vv));
      glVertex3f(cos(vv) * cos(ww) * round_width,
                 cos(vv) * sin(ww) * round_width,
                 sin(vv) * round_width);

      glNormal3f(cos(v) * cos(ww), cos(v) * sin(ww), sin(v));
      glVertex3f(cos(v) * cos(ww) * round_width,
                 cos(v) * sin(ww) * round_width,
                 sin(v) * round_width);
    }
  }

  glEnd();

  if (draw_normals) {
    
    for (int i = 0; i < round_steps; i++) {
      v = i * step_size;

      for (int j = 0; j < round_steps; j++) {
        w = j * step_size;

        DrawArrow(cos(v) * cos(w) * round_width,
                  cos(v) * sin(w) * round_width,
                  sin(v) * round_width,
                  cos(v) * cos(w) * round_width * 1.5,
                  cos(v) * sin(w) * round_width * 1.5,
                  sin(v) * round_width * 1.5);
        
      }
    }
  }

}

void RoundedCube::MakeRoundedSide() {

  float step_size = M_PI/2/round_steps;
  float v, vv;
  
  glBegin(GL_QUADS);
  glColor3f(0.5, 0.5, 0.5);
  
  for (int i = 0; i < round_steps; i++) {
    v = i * step_size;
    vv = (i+1)*step_size;

    glNormal3f(cos(v), sin(v), 0);
    glVertex3f(cos(v) * round_width, sin(v) * round_width,  side_length/2);
    glVertex3f(cos(v) * round_width, sin(v) * round_width, -side_length/2);

    glNormal3f(cos(vv), sin(vv), 0);
    glVertex3f(cos(vv) * round_width, sin(vv) * round_width, -side_length/2);
    glVertex3f(cos(vv) * round_width, sin(vv) * round_width,  side_length/2);
    
  }

  glEnd();
  
  if (draw_normals) {
    for (int i = 0; i < round_steps; i++) {
      v = i * step_size;
      vv = (i+1)*step_size;
      
      DrawArrow(cos(v) * round_width,       sin(v) * round_width,       0,
                cos(v) * round_width * 1.5, sin(v) * round_width * 1.5, 0);

      DrawArrow(cos(vv) * round_width,       sin(vv) * round_width,       0,
                cos(vv) * round_width * 1.5, sin(vv) * round_width * 1.5, 0);
    }
  }

}

void RoundedCube::render() {

  glPushMatrix();

  /* We start with the sides */
  
  glMaterialfv(GL_FRONT, GL_SPECULAR, sides_specular);
  glMaterialfv(GL_FRONT, GL_SHININESS, sides_shininess);

  glBegin(GL_QUADS);

  // The Right
  //glColor3fv(sides[0]);
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, sides[0]);
  glNormal3f(1, 0, 0);

  glVertex3f(side_length/2 + round_width,  side_length/2,  side_length/2);
  glVertex3f(side_length/2 + round_width,  side_length/2, -side_length/2);
  glVertex3f(side_length/2 + round_width, -side_length/2, -side_length/2);
  glVertex3f(side_length/2 + round_width, -side_length/2,  side_length/2);

  // The Top
  //glColor3fv(sides[1]);
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, sides[1]);
  glNormal3f(0, 1, 0);

  glVertex3f(-side_length/2,  side_length/2 + round_width, -side_length/2);
  glVertex3f(-side_length/2,  side_length/2 + round_width,  side_length/2);
  glVertex3f( side_length/2,  side_length/2 + round_width,  side_length/2);
  glVertex3f( side_length/2,  side_length/2 + round_width, -side_length/2);

  // The Front
  //glColor3fv(sides[2]);
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, sides[2]);
  glNormal3f(0, 0, 1);

  glVertex3f( side_length/2,  side_length/2,  side_length/2 + round_width);
  glVertex3f(-side_length/2,  side_length/2,  side_length/2 + round_width);
  glVertex3f(-side_length/2, -side_length/2,  side_length/2 + round_width);
  glVertex3f( side_length/2, -side_length/2,  side_length/2 + round_width);

  // The Back
  //glColor3fv(sides[3]);
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, sides[3]);
  glNormal3f(0, 0, -1);

  glVertex3f( side_length/2,  side_length/2, -side_length/2 - round_width);
  glVertex3f(-side_length/2,  side_length/2, -side_length/2 - round_width);
  glVertex3f(-side_length/2, -side_length/2, -side_length/2 - round_width);
  glVertex3f( side_length/2, -side_length/2, -side_length/2 - round_width);

  // The Bottom
  //glColor3fv(sides[4]);
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, sides[4]);
  glNormal3f(0, -1, 0);

  glVertex3f(-side_length/2, -side_length/2 - round_width, -side_length/2);
  glVertex3f(-side_length/2, -side_length/2 - round_width,  side_length/2);
  glVertex3f( side_length/2, -side_length/2 - round_width,  side_length/2);
  glVertex3f( side_length/2, -side_length/2 - round_width, -side_length/2);

  // The Left
  //glColor3fv(sides[5]);
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, sides[5]);
  glNormal3f(-1, 0, 0);

  glVertex3f(-side_length/2 - round_width,  side_length/2,  side_length/2);
  glVertex3f(-side_length/2 - round_width,  side_length/2, -side_length/2);
  glVertex3f(-side_length/2 - round_width, -side_length/2, -side_length/2);
  glVertex3f(-side_length/2 - round_width, -side_length/2,  side_length/2);

  glEnd();

  if (draw_normals) {
    // Back normals
    DrawArrow(-side_length/2, -side_length/2, -side_length/2-round_width,
              -side_length/2, -side_length/2, -side_length/2-1.5*round_width);
    DrawArrow(-side_length/2,  side_length/2, -side_length/2-round_width,
              -side_length/2,  side_length/2, -side_length/2-1.5*round_width);
    DrawArrow( side_length/2,  side_length/2, -side_length/2-round_width,
               side_length/2,  side_length/2, -side_length/2-1.5*round_width);
    DrawArrow( side_length/2, -side_length/2, -side_length/2-round_width,
               side_length/2, -side_length/2, -side_length/2-1.5*round_width);

    // Front normals
    DrawArrow(-side_length/2, -side_length/2, side_length/2+round_width,
              -side_length/2, -side_length/2, side_length/2+1.5*round_width);
    DrawArrow(-side_length/2,  side_length/2, side_length/2+round_width,
              -side_length/2,  side_length/2, side_length/2+1.5*round_width);
    DrawArrow( side_length/2,  side_length/2, side_length/2+round_width,
               side_length/2,  side_length/2, side_length/2+1.5*round_width);
    DrawArrow( side_length/2, -side_length/2, side_length/2+round_width,
               side_length/2, -side_length/2, side_length/2+1.5*round_width);

    // Top normals
    DrawArrow(-side_length/2, side_length/2+round_width,     -side_length/2,
              -side_length/2, side_length/2+1.5*round_width, -side_length/2);
    DrawArrow(-side_length/2, side_length/2+round_width,      side_length/2,
              -side_length/2, side_length/2+1.5*round_width,  side_length/2);
    DrawArrow( side_length/2, side_length/2+round_width,      side_length/2,
               side_length/2, side_length/2+1.5*round_width,  side_length/2);
    DrawArrow( side_length/2, side_length/2+round_width,     -side_length/2,
               side_length/2, side_length/2+1.5*round_width, -side_length/2);

    // Bottom normals
    DrawArrow(-side_length/2, -side_length/2-round_width,     -side_length/2,
              -side_length/2, -side_length/2-1.5*round_width, -side_length/2);
    DrawArrow(-side_length/2, -side_length/2-round_width,      side_length/2,
              -side_length/2, -side_length/2-1.5*round_width,  side_length/2);
    DrawArrow( side_length/2, -side_length/2-round_width,      side_length/2,
               side_length/2, -side_length/2-1.5*round_width,  side_length/2);
    DrawArrow( side_length/2, -side_length/2-round_width,     -side_length/2,
               side_length/2, -side_length/2-1.5*round_width, -side_length/2);

    // Right normals
    DrawArrow(side_length/2+round_width,     -side_length/2, -side_length/2,
              side_length/2+1.5*round_width, -side_length/2, -side_length/2);
    DrawArrow(side_length/2+round_width,     -side_length/2,  side_length/2,
              side_length/2+1.5*round_width, -side_length/2,  side_length/2);
    DrawArrow(side_length/2+round_width,      side_length/2,  side_length/2,
              side_length/2+1.5*round_width,  side_length/2,  side_length/2);
    DrawArrow(side_length/2+round_width,      side_length/2, -side_length/2,
              side_length/2+1.5*round_width,  side_length/2, -side_length/2);

    // Left normals
    DrawArrow(-side_length/2-round_width,     -side_length/2, -side_length/2,
              -side_length/2-1.5*round_width, -side_length/2, -side_length/2);
    DrawArrow(-side_length/2-round_width,     -side_length/2,  side_length/2,
              -side_length/2-1.5*round_width, -side_length/2,  side_length/2);
    DrawArrow(-side_length/2-round_width,      side_length/2,  side_length/2,
              -side_length/2-1.5*round_width,  side_length/2,  side_length/2);
    DrawArrow(-side_length/2-round_width,      side_length/2, -side_length/2,
              -side_length/2-1.5*round_width,  side_length/2, -side_length/2);



  }
  
  /* Next comes the bevels */

  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, bevel_diffuse);
  glMaterialfv(GL_FRONT, GL_SPECULAR, bevel_specular);
  glMaterialfv(GL_FRONT, GL_SHININESS, bevel_shininess);
  //glColor3f(0.2, 0.2, 0.2);

  // Upper, right, back Corner
  glTranslatef(side_length/2, side_length/2, side_length/2);
  MakeCornerSphere();

  // Upper, left, back Corner
  glTranslatef(-side_length, 0, 0);
  glRotatef(90, 0, 0, 1);
  MakeCornerSphere();
  glRotatef(-90, 0, 0, 1);

  // Lower, left, back Corner
  glTranslatef(0, -side_length, 0);
  glRotatef(180, 0, 0, 1);
  MakeCornerSphere();
  glRotatef(-180, 0, 0, 1);

  // Lower, right, back Corner
  glTranslatef(side_length, 0, 0);
  glRotatef(270, 0, 0, 1);
  MakeCornerSphere();
  glRotatef(-270, 0, 0, 1);

  // Upper, right, front Corner
  glTranslatef(0, side_length, -side_length);
  glRotatef(90, 0, 1, 0);
  MakeCornerSphere();
  glRotatef(-90, 0, 1, 0);


  // Upper, left, front Corner
  glTranslatef(-side_length, 0, 0);
  glRotatef(180, 0, 1, 0);
  MakeCornerSphere();
  glRotatef(-180, 0, 1, 0);

  // Lower, left, front Corner
  glTranslatef(0, -side_length, 0);
  glRotatef(180, 0, 1, 0);
  glRotatef(90, 1, 0, 0);
  MakeCornerSphere();
  glRotatef(-90, 1, 0, 0);
  glRotatef(-180, 0, 1, 0);

  // Lower, right, front Corner
  glTranslatef(side_length, 0, 0);
  glRotatef(180, 1, 0, 0);
  MakeCornerSphere();
  glRotatef(-180, 1, 0, 0);

  glTranslatef(0, side_length, side_length/2);
  
  MakeRoundedSide();
  
  glTranslatef(-side_length, 0, 0);
  glRotatef(90, 0, 0, 1);
  MakeRoundedSide();
  glRotatef(-90, 0, 0, 1);

  glTranslatef(0, -side_length, 0);
  glRotatef(180, 0, 0, 1);
  MakeRoundedSide();
  glRotatef(-180, 0, 0, 1);

  glTranslatef(side_length, 0, 0);
  glRotatef(270, 0, 0, 1);
  MakeRoundedSide();
  glRotatef(-270, 0, 0, 1);

  glTranslatef(0, side_length/2, side_length/2);
  glRotatef(90, 1, 0, 0);
  MakeRoundedSide();

  glTranslatef(-side_length, 0, 0);
  glRotatef(90, 0, 0, 1);
  MakeRoundedSide();

  glTranslatef(-side_length, 0, 0);
  glRotatef(90, 0, 0, 1);
  MakeRoundedSide();

  glTranslatef(-side_length, 0, 0);
  glRotatef(90, 0, 0, 1);
  MakeRoundedSide();

  glRotatef(-90, 1, 0, 0);
  glTranslatef(0, side_length/2, -side_length/2);
  
  MakeRoundedSide();

  glTranslatef(-side_length, 0, 0);
  glRotatef(90, 0, 0, 1);
  MakeRoundedSide();

  glTranslatef(-side_length, 0, 0);
  glRotatef(90, 0, 0, 1);
  MakeRoundedSide();

  glTranslatef(-side_length, 0, 0);
  glRotatef(90, 0, 0, 1);
  MakeRoundedSide();

  glPopMatrix();

}
