// A class that represents a cube with rounded edges.
// Copyright (C) 2002  Martin Geisler <gimpster@gimpster.com>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef ROUNDEDCUBE_H
#define ROUNDEDCUBE_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <GL/glut.h>
#include <math.h>

/** Models a cube with rounded edges.
 *
 *  The cube will be made up of six squares with rounded edges
 *  in-between. The sides will be red, green, blue, orange, yellow,
 *  and white just as the original Rubik's Cube. They will only
 *  reflect little of the specular light that shines on them, but they
 *  will reflect all the diffuse light.
 *
 *  The rounded edges will be black, but will have more intense
 *  highlights than the sides. This gives a nice shiny effect when the
 *  cube rotates in front of a light. */
class RoundedCube {

 private:
  float side_length;
  float round_width;
  int round_steps;

  void CopyColor(float a[], float b[]);
  void RotateColors(int a, int b, int c, int d);

  void MakeCornerSphere();
  void MakeRoundedSide();
  void DrawMarker();
  void DrawArrow(float x1, float y1, float z1,
                 float x2, float y2, float z2);

  float sides[6][3];
  float sides_specular[4];
  float sides_shininess[1];
  float bevel_diffuse[4];
  float bevel_specular[4];
  float bevel_shininess[1];

 public:
  /** The x-, y-, and z-axis. */
  enum axis { Xaxis, Yaxis, Zaxis };

  /** Should normals be drawn?
   *  This should only be set to \c true, if you're debugging. */
  bool draw_normals;

  /** Constructs a cube with rounded edges.
   *
   *  @param s_length the length of the sides. This is only the length
   *  of the squares. To get the total width/height/depth of the cube,
   *  you'll have to add two times \a r_width as well.
   *
   *  @param r_width the width of the roundings.
   *
   *  @param r_steps the number of steps used to do the rounded edges.
   *  The edges will consist of \a r_steps rectangles --- the larger the
   *  number, the finer the edge. */
  RoundedCube(float s_length = 0.8, float r_width = 0.1, int r_steps = 2);

  /** Draws the cube.
   *  The cube will be drawn centered around (0, 0). */
  void render();

  /** Rotates the cube 90 degrees.
   *
   *  The cube will rotate by changing the color of the sides.
   *
   *  @param a the axis to rotate around. The axis are local to the
   *  cube.
   *
   *  @param direction if direction is \c true, then the cube will be
   *  rotated clockwise, otherwise it will be rotated
   *  counter-clockwise. */
  void rotate(axis a, bool direction);

};

#endif
