// A 3D simulation of Rubik's Cube using OpenGL.
// Copyright (C) 2002  Martin Geisler <gimpster@gimpster.com>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "RubiksCube.h"

RubiksCube::RubiksCube() {
  for (int i = 0; i < 3; i++) {
    for (int j = 0; j < 3; j++) {
      for (int k = 0; k < 3; k++) {
        cubes[i][j][k] = new RoundedCube();
      }
    }
  }
}

RubiksCube::~RubiksCube() {
  /* Delete the objects */
  for (int i = 0; i < 3; i++) {
    for (int j = 0; j < 3; j++) {
      for (int k = 0; k < 3; k++) {
        delete cubes[i][j][k];
      }
    }
  }
}

void RubiksCube::render() {

  glRotatef(total_x, 1, 0, 0);
  glRotatef(total_y, 0, 1, 0);
  glRotatef(total_z, 0, 0, 1);

  for (int i = 0; i < 3; i++) {
    glRotatef(rot_x[i], 1, 0, 0);
    for (int j = 0; j < 3; j++) {
      glRotatef(rot_y[j], 0, 1, 0);
      for (int k = 0; k < 3; k++) {
        glRotatef(rot_z[k], 0, 0, 1);

        glTranslatef(i-1, j-1, k-1);
        cubes[i][j][k]->render();
        glTranslatef(1-i, 1-j, 1-k);

        glRotatef(-rot_z[k], 0, 0, 1);
      }
      glRotatef(-rot_y[j], 0, 1, 0);
    }
    glRotatef(-rot_x[i], 1, 0, 0);
  }

  glRotatef(-total_z, 0, 0, 1);
  glRotatef(-total_y, 0, 1, 0);
  glRotatef(-total_x, 1, 0, 0);
}

void RubiksCube::rotateX(unsigned int block, int degrees) {

  if (!(rot_y[0] == rot_y[1] && rot_y[1] == rot_y[2] && rot_y[2] == 0 &&
        rot_z[0] == rot_z[1] && rot_z[1] == rot_z[2] && rot_z[2] == 0)) {
    /* We can't rotate about the x axis, as the other axis aren't
       alligned. */
    return;
  }

  rot_x[block] += degrees;

  if (degrees < 0 && rot_x[block] == -90) {
    /* We stop rotating, and instead rearrange the cubes. */
    rot_x[block] = 0;
    
    RoundedCube *tmpcube = cubes[block][0][0];
    cubes[block][0][0] = cubes[block][2][0];
    cubes[block][2][0] = cubes[block][2][2];
    cubes[block][2][2] = cubes[block][0][2];
    cubes[block][0][2] = tmpcube;
    
    tmpcube = cubes[block][1][0];
    cubes[block][1][0] = cubes[block][2][1];
    cubes[block][2][1] = cubes[block][1][2];
    cubes[block][1][2] = cubes[block][0][1];
    cubes[block][0][1] = tmpcube;
    
    for (int j = 0; j < 3; j++) {
      for (int k = 0; k < 3; k++) {
        cubes[block][j][k]->rotate(RoundedCube::Xaxis, false);
      }
    }
  } else if (degrees > 0 && rot_x[block] == 90) {
    /* We stop rotating, and instead rearrange the cubes. */
    rot_x[block] = 0;
    
    RoundedCube *tmpcube = cubes[block][0][2];
    cubes[block][0][2] = cubes[block][2][2];
    cubes[block][2][2] = cubes[block][2][0];
    cubes[block][2][0] = cubes[block][0][0];
    cubes[block][0][0] = tmpcube;
    
    tmpcube = cubes[block][0][1];
    cubes[block][0][1] = cubes[block][1][2];
    cubes[block][1][2] = cubes[block][2][1];
    cubes[block][2][1] = cubes[block][1][0];
    cubes[block][1][0] = tmpcube;
    
    for (int j = 0; j < 3; j++) {
      for (int k = 0; k < 3; k++) {
        cubes[block][j][k]->rotate(RoundedCube::Xaxis, true);
      }
    }
  }
  
  if ((rot_x[0] == rot_x[1] ||
       rot_x[0] == rot_x[1]-90 ||
       rot_x[0] == rot_x[1]+90) &&
      (rot_x[1] == rot_x[2] ||
       rot_x[1] == rot_x[2]-90 ||
       rot_x[1] == rot_x[2]+90)) {
    /* All three blocks have been turned the same amount, so we reset
       rot_x. */
    total_x += rot_x[0];
    rot_x[0] = rot_x[1] = rot_x[2] = 0;
  }

}

void RubiksCube::rotateY(unsigned int block, int degrees) {

  if (!(rot_x[0] == rot_x[1] && rot_x[1] == rot_x[2] && rot_x[2] == 0 &&
        rot_z[0] == rot_z[1] && rot_z[1] == rot_z[2] && rot_z[2] == 0)) {
    /* We can't rotate about the y axis, as the other axis aren't
       alligned. */
    return;
  }
  
  rot_y[block] += degrees;
  
  if (degrees < 0 && rot_y[block] == -90) {
    /* We stop rotating, and instead rearrange the cubes. */
    rot_y[block] = 0;
    
    RoundedCube *tmpcube = cubes[0][block][2];
    cubes[0][block][2] = cubes[2][block][2];
    cubes[2][block][2] = cubes[2][block][0];
    cubes[2][block][0] = cubes[0][block][0];
    cubes[0][block][0] = tmpcube;
                                             
    tmpcube = cubes[0][block][1];
    cubes[0][block][1] = cubes[1][block][2];
    cubes[1][block][2] = cubes[2][block][1];
    cubes[2][block][1] = cubes[1][block][0];
    cubes[1][block][0] = tmpcube;
    
    for (int i = 0; i < 3; i++) {
      for (int k = 0; k < 3; k++) {
        cubes[i][block][k]->rotate(RoundedCube::Yaxis, false);
      }
    }
  } else if (degrees > 0 && rot_y[block] == 90) {
    /* We stop rotating, and instead rearrange the cubes. */
    rot_y[block] = 0;
    
    RoundedCube *tmpcube = cubes[0][block][0];
    cubes[0][block][0] = cubes[2][block][0];
    cubes[2][block][0] = cubes[2][block][2];
    cubes[2][block][2] = cubes[0][block][2];
    cubes[0][block][2] = tmpcube;
    
    tmpcube = cubes[1][block][0];
    cubes[1][block][0] = cubes[2][block][1];
    cubes[2][block][1] = cubes[1][block][2];
    cubes[1][block][2] = cubes[0][block][1];
    cubes[0][block][1] = tmpcube;
    
    for (int i = 0; i < 3; i++) {
      for (int k = 0; k < 3; k++) {
        cubes[i][block][k]->rotate(RoundedCube::Yaxis, true);
      }
    }
  }

  if ((rot_y[0] == rot_y[1] ||
       rot_y[0] == rot_y[1]-90 ||
       rot_y[0] == rot_y[1]+90) &&
      (rot_y[1] == rot_y[2] ||
       rot_y[1] == rot_y[2]-90 ||
       rot_y[1] == rot_y[2]+90)) {
    /* All three blocks have been turned the same amount, so we reset
       rot_y. */
    total_y += rot_y[0];
    rot_y[0] = rot_y[1] = rot_y[2] = 0;
  }

}

void RubiksCube::rotateZ(unsigned int block, int degrees) {

  if (!(rot_x[0] == rot_x[1] && rot_x[1] == rot_x[2] && rot_x[2] == 0 &&
        rot_y[0] == rot_y[1] && rot_y[1] == rot_y[2] && rot_y[2] == 0)) {
    /* We can't rotate about the z axis, as the other axis aren't
       alligned. */
    return;
  }

  rot_z[block] += degrees;

  if (degrees < 0 && rot_z[block] == -90) {
    /* We stop rotating, and instead rearrange the cubes. */
    rot_z[block] = 0;
    
    RoundedCube *tmpcube = cubes[0][0][block];
    cubes[0][0][block]   = cubes[2][0][block];
    cubes[2][0][block]   = cubes[2][2][block];
    cubes[2][2][block]   = cubes[0][2][block];
    cubes[0][2][block]   = tmpcube;
    
    tmpcube            = cubes[1][0][block];
    cubes[1][0][block] = cubes[2][1][block];
    cubes[2][1][block] = cubes[1][2][block];
    cubes[1][2][block] = cubes[0][1][block];
    cubes[0][1][block] = tmpcube;
    
    for (int i = 0; i < 3; i++) {
      for (int j = 0; j < 3; j++) {
        cubes[i][j][block]->rotate(RoundedCube::Zaxis, false);
      }
    }
  } else if (degrees > 0 && rot_z[block] == 90) {
    /* We stop rotating, and instead rearrange the cubes. */
    rot_z[block] = 0;
    
    RoundedCube *tmpcube = cubes[0][2][block];
    cubes[0][2][block] = cubes[2][2][block];
    cubes[2][2][block] = cubes[2][0][block];
    cubes[2][0][block] = cubes[0][0][block];
    cubes[0][0][block] = tmpcube;
                                             
    tmpcube = cubes[0][1][block];
    cubes[0][1][block] = cubes[1][2][block];
    cubes[1][2][block] = cubes[2][1][block];
    cubes[2][1][block] = cubes[1][0][block];
    cubes[1][0][block] = tmpcube;
    
    for (int i = 0; i < 3; i++) {
      for (int j = 0; j < 3; j++) {
        cubes[i][j][block]->rotate(RoundedCube::Zaxis, true);
      }
    }
  }

  if ((rot_z[0] == rot_z[1] ||
       rot_z[0] == rot_z[1]-90 ||
       rot_z[0] == rot_z[1]+90) &&
      (rot_z[1] == rot_z[2] ||
       rot_z[1] == rot_z[2]-90 ||
       rot_z[1] == rot_z[2]+90)) {
    /* All three blocks have been turned the same amount, so we reset
       rot_z. */
    total_z += rot_z[0];
    rot_z[0] = rot_z[1] = rot_z[2] = 0;
  }

}
