// A -*- C++ -*- class for parsing command line options.
//
// Copyright (C) 2001  Ole Laursen <olau@hardworking.dk>
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330,
// Boston, MA  02111-1307  USA.


#ifndef COMMANDOPTIONS_HEADER
#define COMMANDOPTIONS_HEADER

#include <vector>
#include <string>
#include <exception>
#include <sstream>

/** A class for parsing command line options. By using this class, you
    wont have to redo the tedious task of making a command line parser
    every time you need one.

    The command line is parsed and checked for errors. An exception
    will be thrown in any of these circumstances:
    - an unrecognizable option or flag was encountered.
    - the argument to an option could not be parsed. This means that
      an exception is thrown, if the type of the argument doesn't
      match the type of the variable bound to the argument. This is
      determined by using the \c << operator.
    - too many or too few arguments are given.
    - several options are put into one group. The rule is, that there
      can only be one option per group, as each option requires an
      argument. If \c -f is an option that requires a filename (most
      likely a \c string ), and \c -n also is an option, then the
      following command line will make \c commandoptions throw an
      exception: \code $ my_program -fn filename \endcode
    - an option is used without an argument.
    - a single dash is found without a trailing character. Also, if
      two dashes are found without the name of a long option or long
      flag. This means that \c commandoptions currently \e doesn't
      follow the GNU style, which treats everything following two
      consecutive dashes arguments.

    Here is a little program that demonstrates the use of \c commandoptions:
    \dontinclude commandoptions-test.cpp

    We start by including the standard IO stream library (which is
    needed by our program and not by \c commandoptions ) and the \c
    commandoptions header:
    \skip include
    \until commandoptions
    
    We can now start our \c main function and decleare our \c
    commandoptions object:
    \until commandoptions
    
    Our little program has a number of variables that the user can
    control. These variables are declared next:
    \until victim
    
    Now that the variables have been declared and initialised, we can
    register them with our \c commandoptions object:
    \until argument
    
    We can now try to parse the command line. Since this might fail,
    we use a \c try block:
    \until argv
    
    If the parsing fails, a \c commandoptions_error object will be
    thrown, and we will end up in the \c catch block. The error-object
    has a \c commandoptions_error::what method that we use to get an
    explaination of the error:
    \until what
    
    The program should now terminate with a non-zero exitcode to
    indicate that an error has occured:
    \until }
    
    If everything went well, then we can start using the variables we
    declared previously. They will be updated to reflect the arguments
    given by the user on the command line. The rest of the code is:
    \until EOF

 */
class commandoptions
{
public:
  /** Register an option with a reference to a variable. The variable will be
      updated with the argument on the command line.
      \param par the variable to update. The variable is tied to the option,
             and it will be updated when process_command_line is called.
      \param long_name the long name of the option. The user can then enable
             the option by invoking the program with \c --long_name. The long
             name will be part of the help displayed when the user invokes the
             program with \c --help as an option.
      \param short_name the short name for the option. This is a single \c char.
             The user can then enable the option by invoking the program with
             \c -short_name. The short name will be part of the help displayed
             when the program is invoked with the \c --help option.
      \param des the description of the option. The description will be part of
             the help displayed when the program in invoed with the \c --help
             option. It will be placed after the long and short name, so it
             should be no longer than one line.
      \param arg_name the name of the option. This will be part of the help
             displayed when the program is invoked with the \c -h or \c -?
             options.
  */
  // register an option with pointer to variable, long and short argument names, a
  // description of the option and the name of the argument to the option
  template <typename T>
  void register_option(T &par, std::string long_name, char short_name,
		       std::string des, std::string arg_name)
  {
    option_table.push_back(option(new updater<T>(par), long_name, short_name, des, arg_name));
  }

  /** Register a flag with a reference to a \c bool variable. The variable will
      be inverted if the flag is found on the command line. This means that you
      should give the variable a suitable default value before you call
      process_command_line. If the flag is present more than one time on the
      command line, it will be parsed as if only occured \e once.
      \param par the variable to update. The variable is tied to the flag,
             and it will be updated when process_command_line is called.
      \param long_name the long name of the flag. The user can then toggle
             the flag by invoking the program with \c --long_name. The long
             name will be part of the help displayed when the user invokes the
             program with \c --help as an option.
      \param short_name the short name for the flag. This is a single \c char.
             The user can then toggle the flag by invoking the program with
             \c -short_name. The short name will be part of the help displayed
             when the program is invoked with the \c --help option.
      \param des the description of the flag. The description will be part of
             the help displayed when the program in invoed with the \c --help
             option. It will be placed after the long and short name, so it
             should be no longer than one line.
  */
  void register_flag(bool &par, std::string long_name, char short_name, std::string des)
  {
    flag_table.push_back(flag(par, long_name, short_name, des));
  }

  /** Register a mandatory argument with a reference to a variable. The
      variable will be updated with the value found on the command line.
      \param par the variable to update. The variable is tied to the flag,
             and it will be updated when process_command_line is called. This
             is done with the help of \c >> which means that you can make your
             program accept a vector as an argument, if you have implemented the
             \c operator>> method in your vectorclass.
      \param des the description of the argument. The description will be part
             of the help displayed when the program in invoed with the
             \c --help option. It will be placed after the long and short name,
             so it should be no longer than one line.
  */
  template <typename T>
  void register_argument(T &par, std::string name, std::string des)
  {
    argument_table.push_back(argument(new updater<T>(par), name, des));
  }

  /** Process the arguments, options, and flags on the command line. You should
      call this method after you have registered your arguments, options, and
      flags with the object.
      \sa register_argument, register_option, and register_flag.
  */
  void process_command_line(int argc, const char *argv[]);

  /** Cleans dynamically allocated stuff up. */
  ~commandoptions()
  {
    // make sure dynamically allocated stuff is cleaned up
    for (std::vector<updaterbase *>::iterator it = clean_up_table.begin();
	 it != clean_up_table.end(); ++it)
      delete (*it);
  }


private:
  // member helpers

  /** Formats a string for use a required argument in the help.
      \param s the string.
      \return the string enclosed in angle-brackets: \c <s>. */
  std::string requirify(std::string s)
  {
    return '<' + s + '>';
  }

  /** Formats a string for use a optional argument in the help.
      \param s the string.
      \return the string enclosed in brackets: \c [s]. */
  std::string optionalify(std::string s)
  {
    return '[' + s + ']';
  }

  /** Strips the path from a filename.
      \param s the filename.
      \return the basename of \a s. */
  std::string strip_path(std::string s);

  /** Prings terse usage instructions.
      \param executable the name of the executable. */
  void print_usage(std::string executable);
  /** Prints detailed help.
      \param executable the name of the executable. */
  void print_help(std::string executable);

  // private data structures
  // first a few classes for updating a variable from a stream
  class updaterbase
  {
  public:
    virtual bool update(std::string s) = 0;
  };

  template <typename T>
  class updater: public updaterbase
  {
  public:
    updater(T &par):
      var(par)
    {}

    // update the contained variable through reference - return false upon failure
    virtual bool update(std::string s)
    {
      std::istringstream is(s);
      is >> var;
      if (!is)
	return false;
      else
	return true;
    }
    
  private:
    T &var;
  };

  // a structure for keeping a command line option
  struct option 
  {
    updaterbase *par;
    std::string long_name, des, arg_name;
    char short_name;

    option(updaterbase *p, std::string ln, char sn, std::string d, std::string an):
      par(p), long_name(ln), des(d), arg_name(an), short_name(sn)
    {}
    option(const option &o):
      par(o.par), long_name(o.long_name), des(o.des), arg_name(o.arg_name), short_name(o.short_name)
    {}
  };

  // one for keeping a flag
  struct flag
  {
    bool *par;
    std::string long_name, des;
    char short_name;
    
    flag(bool &p, std::string ln, char sn, std::string d):
      par(&p), long_name(ln), des(d), short_name(sn)
    {}
  };
  
  // and one for keeping an argument
  struct argument
  {
    updaterbase *par;
    std::string name, des;
    
    argument(updaterbase *p, std::string n, std::string d):
      par(p), name(n), des(d)
    {}
    argument(const argument &a):
      par(a.par), name(a.name), des(a.des)
    {}
  };

  /** Table of options. */
  std::vector<option> option_table;
  /** Table of flags. */
  std::vector<flag> flag_table;
  /** Table of arguments. */
  std::vector<argument> argument_table;
  /** Shorthand for an optionsiterator. */
  typedef std::vector<option>::iterator option_iterator;
  /** Shorthand for a flagiterator. */
  typedef std::vector<flag>::iterator flag_iterator;
  /** Shorthand for an argumentiterator. */
  typedef std::vector<argument>::iterator argument_iterator;

  std::vector<updaterbase *> clean_up_table;
};

template <> class commandoptions::updater<std::string>: public commandoptions::updaterbase
{
public:
  updater(std::string &par):
    var(par)
  {}

  // update the contained variable through reference - return false upon failure
  virtual bool update(std::string s)
  {
    var = s;
    return true;
  }
    
private:
  std::string &var;
};
  
/** An exception class. */
// an exception class
class commandoptions_error: std::exception
{
public:
  /** Constructs a new error.
      \param s the errormessage. */
  commandoptions_error(std::string s) throw(): msg(s)
  {}
  /** Does nothing? */
  ~commandoptions_error() throw()
  {}

  /** Returns the errormessage. */
  virtual const char *what() const throw()
  {
    return msg.c_str();
  }

private:
  std::string msg;
};


#endif // COMMANDOPTIONS_HEADER
